@@ -2,7 +2,7 @@
  **
  ** Atari++ emulator (c) 2002 THOR-Software, Thomas Richter
  **
- ** $Id: antic.hpp,v 1.64 2013-01-02 12:04:12 thor Exp $
+ ** $Id: antic.hpp,v 1.71 2021/08/16 10:31:01 thor Exp $
  **
  ** In this module: Antic graphics emulation
  **
@@ -71,9 +71,6 @@ public:
   static const int WindowHeight        INIT(DisplayHeight - DisplayStart);
   // Total number of rows visible.
   static const int WindowWidth         INIT(DisplayWidth - 32);
-#ifndef HAS_PRIVATE_ACCESS
-public:
-#endif
   //
   // DMA slot allocation models for various ANTIC DMA activities:
   // Pre-allocated DMA slots for memory refresh
@@ -93,10 +90,6 @@ public:
   // Pre-allocated DMA slots for playfield graphics requiring at most 48 bytes data
   static const UBYTE Playfield48Fetch[103];
   //
-#ifdef HAS_PRIVATE_ACCESS
-private:
-#endif
-  //
   // Pointer to the CPU for the computations
   class CPU        *Cpu;
   //
@@ -140,7 +133,7 @@ public:
   struct CharacterGenerator {
     // Character generator: Orientation in vertical direction: +1 or -1
     class AdrSpace *Ram;          // where to fetch characters from
-    bool            UpsideDown;   // if true, then characters are displayed upside down
+    UBYTE           UpsideDown;   // if 7 then characters are displayed upside down, otherwise 0
     int             InvertMask;   // if code & this is non-null, then it gets inverted
     int             BlankMask;    // if code & this is non-null, then it gets blanked out
     ADR             CharBase;     // Character base
@@ -167,11 +160,10 @@ private:
   UBYTE NMIEnable;    // NMI masking/enable register
   UBYTE NMIStat;      // NMI status register
   UBYTE DMACtrl;      // DMA control register
-  UBYTE DMACtrlShadow;// DMA control register, updated at the end of the scanline
   UBYTE CharCtrl;     // Character control register shaddow of the character generator
   UBYTE HScroll;      // Horizontal scroll offset
   UBYTE VScroll;      // Vertical scroll offset
-  UBYTE BusNoise[5];  // latest BusNoise rolled out at the start of each mode-line
+  UBYTE PlayerData[5];// Antic service for GTIA: Data fetched or seen as p/m data.
   //
   //
   // Antic intermediate scanline buffer. Data is DMA'd from memory into
@@ -336,16 +328,74 @@ private:
   // mode type.
   struct ModeLine *ModeLines[16];
   // 
-  // Last modeline data. We keep them here as to be able to rebuild the modeline
-  // contents on the fly should the user change the antic settings during the
-  // horizontal line.
-  struct ModeLine *CurrentMode; // current mode of the scan line
+  // Scanline generator: This helper is responsible for drawing a single
+  // scan line.
+  struct Scanline {
+    //
+    // Last modeline data. We keep them here as to be able to rebuild the modeline
+    // contents on the fly should the user change the antic settings during the
+    // horizontal line.
+    struct ModeLine *CurrentMode; // current mode of the scan line
+    //
+    // Output buffer position
+    UBYTE           *LineBuffer;
+    //
+    // Fill-in position in the antic output buffer.
+    UBYTE           *FillIn;
+    //
+    // Minimum and maximum border position
+    int              XMin;
+    int              XMax;
+    int              Width;
+    //
+    // Generated line (for the character modes) of the mode line
+    int              DisplayLine;
+    //
+    // The constructor: Just reset the entries.
+    Scanline(void)
+      : CurrentMode(NULL), LineBuffer(NULL), FillIn(NULL)
+    {
+    }
+    //
+    // Precompute the parameters for generating a single
+    // scan line. First argument is the modeline,
+    // then the active DMA settings for the scrolled screen,
+    // followed by the regular parameters defining the nominal
+    // borders where we clip.
+    // Last are the output buffer itself, followed by the scroll
+    // offset and the current line in the modeline.
+    void ComputeLineParameters(struct ModeLine *mode,
+			       struct DMAGenerator *dma,
+			       struct DMAGenerator *borders,
+			       UBYTE *buffer,int xscroll,int displayline);
+    //
+    // Reset the mode line, no current mode active.
+    void NoMode(void)
+    {
+      CurrentMode = NULL;
+      LineBuffer  = NULL;
+    }
+    //
+    // Generate a single scan line.
+    void GenerateScanline(void) const;
+    //
+    // Check whether the currently active mode is a fiddled mode, i.e. 
+    // a hires mode generating half color clocks.
+    bool isFiddled(void) const
+    {
+      if (CurrentMode)
+	return CurrentMode->Fiddling;
+      return false;
+    }
+    //
+  }      ScanlineGenerator;
   //
   // Current instruction - stored for next line.
   UBYTE  PreviousIR;
   //
   // Some Antic Preferences
   bool NTSC;                 // true if this is an NTSC antic
+  bool isAuto;               // true if video mode comes from the machine
   //
   LONG GTIAStart;            // Horizontal position where GTIA processing starts.
   LONG YPosIncSlot;          // horizontal position where YPos gets incremented
@@ -356,24 +406,8 @@ private:
   // Load data from the playfield into the scanline buffer
   void FetchPlayfield(struct ModeLine *mode,struct DMAGenerator *dma);
   //
-  // Antic::ScanLine (the complex and touchy one)
-  // Generates one scan line of a mode line
-  // screen is the start of the target scanline and has to be bumped by this function.
-  // Generate one scanline of the Antic display
-  // If nmi is true, then a DLI has to generated.
-  // modeline is the mode line generator for this scanline.
-  // fillin is where the playfield data shall go. It may be offset to screen.
-  // width is the number of pixels the modeline shall fill in, shift the
-  // shift offset. 
-  // displayline is the number of the scanline within the modeline this method
-  // generates.
-  // Emulation here starts at value zero of the "horizontal register" as defined
-  // by the technical manual of Antic, sheet 5. The horizontal register itself
-  // counts half color clocks, similar to to our pixel based emulation.
-  // Measurements show that the DLI reaches the CPU about 8 cycles after the
-  // STA WSYNC position at hpos = 208. That fits well to the hypothesis that
-  // a NMI is generated at hpos = 0 since we have 228 half color clocks.
-  void Scanline(struct ModeLine *mode,UBYTE *fillin,int width,int scroll,int displayline);
+  // Fetch the player-missile graphics
+  void FetchPlayerMissiles(void);
   //
   // Generate a complete modeline.
   void Modeline(int ir,int vscroll,int nlines,struct ModeLine *gen);
@@ -424,41 +458,6 @@ public:
   // Start the special Reset Key NMI that is only available at the Atari800 and Atari400.
   void ResetNMI(void);
   //
-  // Private for GTIA: Read Player DMA data
-  void PlayerDMAChannel(int player,int delay,UBYTE &target)
-  {
-    if (DMACtrlShadow & 0x08) {
-      target = UBYTE(Ram->ReadByte(PMActive->PlayerBase[player]
-				   + ((YPos - delay) >> PMActive->YPosShift)));
-    } else {
-      // Interesting side case: If Antic DMA is off, then GTIA does not fetch data
-      // from the bus (Test: Basic BUNDES.BAS)
-      if (DMACtrlShadow & 0x20) {
-	// Otherwise: Return random bus noise
-	target = BusNoise[player];
-      }
-    }
-  }
-  //
-  // Private for GTIA: Read Missile DMA data
-  void MissileDMAChannel(int delay,UBYTE &target)
-  {
-    if (DMACtrlShadow & 0x0c) {
-      // Missile DMA is turned on if player DMA is available.
-      // Test case: POP-Demo, graphics part. (POP.BAS)
-      target = UBYTE(Ram->ReadByte(PMActive->MissileBase
-				   + (YPos >> PMActive->YPosShift) 
-				   - delay));
-    } else {
-      // Interesting side case: If Antic DMA is off, then GTIA does not fetch data
-      // from the bus (Test: Basic BUNDES.BAS)
-      if (DMACtrlShadow & 0x20) {
-	// random bus noise if no DMA channel allocated for it
-	target = BusNoise[4];
-      }
-    }
-  }
-  //
   // Return the current YPos
   int CurrentYPos(void) 
   {
