@@ -2,7 +2,7 @@
  **
  ** Atari++ emulator (c) 2002 THOR-Software, Thomas Richter
  **
- ** $Id: menu.cpp,v 1.69 2020/07/18 15:20:39 thor Exp $
+ ** $Id: menu.cpp,v 1.70 2022/12/20 18:01:33 thor Exp $
  **
  ** In this module: Definition of a graphical frontend with the build-in graphics
  **********************************************************************************/
@@ -786,12 +786,6 @@ Menu::ControlTopic::ControlTopic(class RenderPort *rp,const char *loadname,const
 }
 ///
 
-/// Menu::ControlTopic::~ControlTopic
-Menu::ControlTopic::~ControlTopic(void)
-{
-}
-///
-
 /// Menu::ControlTopic::CreateOptionGadgets
 // This is currently still undefined and does nothing. Later.
 void Menu::ControlTopic::CreateOptionGadgets(List<class Gadget> &GList)
