@@ -2,7 +2,7 @@
  **
  ** Atari++ emulator (c) 2002 THOR-Software, Thomas Richter
  **
- ** $Id: monitor.cpp,v 1.87 2013-01-01 02:29:21 thor Exp $
+ ** $Id: monitor.cpp,v 1.111 2022/12/20 18:01:33 thor Exp $
  **
  ** In this module: Definition of the built-in monitor
  **********************************************************************************/
@@ -27,6 +27,15 @@
 #include <ctype.h>
 ///
 
+/// Defines
+// The OS/2 implementation of ncurses doesn't seem handle the switch from
+// the SDL window well.
+// So for now, we deactivate it.
+#if defined(USE_CURSES) && defined(OS2)
+#undef USE_CURSES
+#endif
+///
+
 /// Statics
 #ifndef HAS_MEMBER_INIT
 const int Monitor::History::MaxHistorySize = 32;
@@ -40,7 +49,8 @@ Monitor::Monitor(class Machine *mach)
   : machine(mach), cpu(mach->CPU()), MMU(mach->MMU()),
     cpuspace(MMU->CPURAM()), anticspace(MMU->AnticRAM()), currentadr(MMU->CPURAM()),
     debugspace(MMU->DebugRAM()),
-    tracefile(NULL), curses(NULL), cmdline(NULL), abort(false), fetchtrace(false),
+    tracefile(NULL), symboltable(NULL), curses(NULL), cmdline(NULL), 
+    abort(false), fetchtrace(false),
     cmdchain(NULL),
     Envi(this,"ENVI","V","environment settings"),
     Splt(this,"SPLT","/","split off display"),
@@ -63,6 +73,7 @@ Monitor::Monitor(class Machine *mach)
     Dump(this,"DUMP","D","[expr]         : display memory contents"),
     SkTb(this,"SKTB","K","[expr]         : stack traceback"),
     Disk(this,"DISK","I","file addr size : read or write memory to a file"),
+    Prof(this,"PROF","O","profile code"),
     // help must go last into the chain to be the first getting displayed.
     Help(this,"HELP","?","display this text")
 { }
@@ -71,12 +82,298 @@ Monitor::Monitor(class Machine *mach)
 /// Monitor::~Monitor
 Monitor::~Monitor(void)
 {
+  ClearSymbolTable();
   if (tracefile)
     fclose(tracefile);
   delete[] cmdline;
 }
 ///
 
+/// Monitor::ClearSymbolTable
+// Remove all known labels from the symbol table.
+void Monitor::ClearSymbolTable(void)
+{
+  struct Symbol *s;
+
+  while(symboltable) {
+    s = symboltable;
+    symboltable = symboltable->next;
+    delete s;
+  }
+}
+///
+
+/// Monitor::ParseSymbolTable
+// This parses a ld65 debug file which contains
+// all the symbols and labels of an assembled and linked ca65
+// binary.
+bool Monitor::ParseSymbolTable(const char *filename)
+{
+  FILE *symbols;
+  char line[256];
+  bool result = false;
+
+  errno   = 0;
+  symbols = fopen(filename,"r");
+  if (symbols) {
+    try {
+      errno = 0;
+      while(!feof(symbols)) {
+	if (fgets(line,sizeof(line),symbols)) {
+	  if (strncmp(line,"sym ",4) == 0 || strncmp(line,"sym\t",4) == 0) {
+	    char *linebuffer;
+	    struct Symbol s;
+	    // This is a symbol line. These are the only ones we care about.
+	    linebuffer = &line[4];
+	    while(*linebuffer == ' ' || *linebuffer == '\t')
+	      linebuffer++;
+	    if (s.ParseLabel(linebuffer)) {
+	      struct Symbol *add;
+	      result = true;
+	      if ((add = const_cast<struct Symbol *>(Symbol::FindSymbol(symboltable,s.name,0,s.type,s.size)))) {
+		// Label already exists, replace its value.
+		add->value  = s.value;
+	      } else {
+		add         = new struct Symbol(s);
+		add->next   = symboltable;
+		symboltable = add;
+	      }
+	    }
+	  }
+	} else if (errno) {
+	  Print("Error %s reading the symbol file %s\n",strerror(errno),filename);
+	}
+      }
+      fclose(symbols);
+    } catch(...) {
+      ClearSymbolTable();
+      fclose(symbols);
+      Print("Cannot allocate symbol table, no memory left.\n");
+    }
+  } else {
+    Print("Error %s opening the symbol file %s\n",strerror(errno),filename);
+  }
+
+  return result;
+}
+///
+
+/// Monitor::Symbol::ParseLabel
+// Parse a label from a parsed-off ld65 debug output line.
+// return true on success.
+bool Monitor::Symbol::ParseLabel(char *line)
+{
+  char buf[32];
+  size_t offs = 0;
+  bool havetype = false;
+  bool havesize = false;
+  bool havevalue = false;
+
+  while(*line == ' ' || *line == '\t')
+    line++;
+  //
+  // Next is the label name in quotes.
+  if (*line != '\"')
+    return false;
+  line++;
+  while(*line != '"') {
+    if (*line == '\\') {
+      line++;
+      if (*line == '"') {
+	if (offs < sizeof(name)-1)
+	  name[offs++] = '\"';
+      } else if (*line == '\\') {
+	if (offs < sizeof(name)-1)
+	  name[offs++] = '\\';
+      } else return false;
+    } else {
+      if (offs < sizeof(name)-1)
+	name[offs++] = *line;
+      line++;
+    }
+  }
+  if (offs == 0)
+    return false;
+  // Name is now complete.
+  name[offs] = 0;
+  //
+  // Names that start with a dot are local labels or assembler
+  // generated labels. They do not become part of the database.
+  if (name[0] == '.')
+    return false;
+  //
+  // Skip over the quote.
+  line++;
+  //
+  // Now parse off the remaining parameters.
+  do {
+    while(*line == ' ' || *line == '\t')
+      line++;
+    if (*line != ',') {
+      if (*line == '\n' && havetype && havesize && havevalue)
+	break;
+      return false;
+    }
+    // Skip over the comma.
+    line++;
+    while(*line == ' ' || *line == '\t')
+      line++;
+    offs = 0;
+    while(*line && *line != '=') {
+      if (offs > sizeof(buf)-1)
+	return false;
+      buf[offs++] = *line++;
+    }
+    buf[offs] = 0;
+    line++; // skip over the equates sign
+    if (!strcmp(buf,"value")) {
+      char *endptr = line;
+      long v = strtol(line,&endptr,0);
+      if (endptr == line)
+	return false;
+      if (v > 0xffff || v < 0)
+	return false;
+      value = UWORD(v);
+      line  = endptr;
+      if (havevalue)
+	return false;
+      havevalue = true;
+    } else if (!strcmp(buf,"addrsize")) {
+      if (!strncmp(line,"absolute",8)) {
+	size  = Absolute;
+	line += 8;
+      } else if (!strncmp(line,"zeropage",8)) {
+	size  = ZeroPage;
+	line += 8;
+      } else return false;
+      if (havesize)
+	return false;
+      havesize = true;
+    } else if (!strcmp(buf,"type")) {
+      if (!strncmp(line,"equate",6)) {
+	type  = Equate;
+	line += 6;
+      } else if (!strncmp(line,"label",5)) {
+	type  = Label;
+	line += 5;
+      } else return false;
+      if (havetype)
+	return false;
+      havetype = true;
+    }
+  } while(true);
+  //
+  return true;
+}
+///
+
+/// Monitor::Symbol::FindSymbol
+// Find a label by its address, size and type.
+const struct Monitor::Symbol *Monitor::Symbol::FindSymbol(const struct Symbol *list,UWORD address,Type t,Size s)
+{
+  const struct Symbol *bestmatch = NULL;
+  int score     = 0;
+  int bestscore = 0;
+  
+  for(;list;list = list->next) {
+    if (list->value == address) {
+      score = 0;
+      switch(t) {
+      case Equate:
+	if (list->type == Equate)
+	  score += 10;
+	else
+	  continue;
+	break;
+      case Label:
+	if (list->type == Label)
+	  score += 10;
+	else
+	  continue;
+	break;
+      case Any:
+	score += 5;
+	break;
+      case PreferLabel:
+	if (list->type == Label)
+	  score += 5;
+	score += 2;
+	break;
+      case PreferEquate:
+	if (list->type == Equate)
+	  score += 5;
+	score += 2;
+	break;
+      }
+      switch(s) {
+      case ZeroPage:
+	if (list->size == ZeroPage)
+	  score += 10;
+	else
+	  continue;
+	break;
+      case Absolute:
+	if (list->size == Absolute)
+	  score += 10;
+	else
+	  continue;
+	break;
+      case All:
+	score += 5;
+	break;
+      case PreferZeroPage:
+	if (list->size == ZeroPage)
+	  score += 5;
+	score += 2;
+	break;
+      case PreferAbsolute:
+	if (list->size == Absolute)
+	  score += 5;
+	score += 2;
+	break;
+      }
+      if (score >= bestscore) {
+	bestscore = score;
+	bestmatch = list;
+      }
+    }
+  }
+  
+  return bestmatch;
+}
+///
+
+/// Monitor::Symbol::FindSymbol
+// Find a symbol by its name, size and type.
+const struct Monitor::Symbol *Monitor::Symbol::FindSymbol(const struct Symbol *list,const char *name,UWORD pc,Type t,Size s)
+{
+  int bestdist = 0xffff;
+  const struct Symbol *bestmatch = NULL;
+  size_t l = strlen(name) + 1;
+
+  if (l > Symbol::MaxLabelSize)
+    l = Symbol::MaxLabelSize;
+
+  while(list) {
+    if ((t == Any || list->type == t) && (s == All || list->size == s)) {
+      
+      if (!strncasecmp(name,list->name,l)) {
+	int dist = int(list->value) - int(pc);
+	if (dist < 0)
+	  dist = -dist;
+	if (dist < bestdist) {
+	  bestdist  = dist;
+	  bestmatch = list;
+	}
+      }
+    }
+    list = list->next;
+  }
+
+  return bestmatch;
+}
+///
+
 /// Monitor::CommandChain
 // Add a command to the command chain
 struct Monitor::Command* &Monitor::CommandChain(void)
@@ -263,7 +560,7 @@ LONG Monitor::EvaluateShift(char *&s)
 LONG Monitor::EvaluateNumeric(char *&s)
 {
   LONG v1;
-
+  
   if (*s == '-') {
     s++;
     v1 = EvaluateNumeric(s);
@@ -302,19 +599,19 @@ LONG Monitor::EvaluateNumeric(char *&s)
   } else if (!memcmp(s,"pc",2)) {
     s += 2;
     return cpu->PC();
-  } else if (*s == 'x') {
+  } else if (*s == 'x' && (!isalnum(s[1]))) {
     s++;
     return cpu->X();
-  } else if (*s == 'y') {
+  } else if (*s == 'y' && (!isalnum(s[1]))) {
     s++;
     return cpu->Y();
-  } else if (*s == 'p') {
+  } else if (*s == 'p' && (!isalnum(s[1]))) {
     s++;
     return cpu->P();
-  } else if (*s == 's') {
+  } else if (*s == 's' && (!isalnum(s[1]))) {
     s++;
     return cpu->S();
-  } else if (*s == 'a' && (!isxdigit(s[1]))) {
+  } else if (*s == 'a'  && (!isalnum(s[1]))) {
     s++;
     return cpu->A();
   } else if (*s == '#') {
@@ -327,10 +624,37 @@ LONG Monitor::EvaluateNumeric(char *&s)
       throw NumericException(this,"Error: Invalid token %s\n",s);
     s  = end;
     return v1;
-  } else {
+  } else if (*s == '$') {
+    char *end;
+    s++;
+    v1 = strtol(s,&end,16);
+    if (errno == ERANGE)
+      throw NumericException(this,"Error: %s is out of range\n",s);
+    if (s == end)
+      throw NumericException(this,"Error: Invalid token %s\n",s);
+    s  = end;
+    return v1;
+  } else if (isalpha(*s)) {
+    char labelname[64]; 
+    const struct Symbol *symbol;
+    char *sptr = s;
+    char *lptr = labelname;
+    while(lptr < labelname + sizeof(labelname) - 1) {
+      *lptr = *sptr;
+      lptr++,sptr++;
+      if (!isalnum(*sptr)) 
+	break;
+    }
+    *lptr  = 0;
+    symbol = Symbol::FindSymbol(symboltable,labelname,cpu->PC(),Symbol::Any,Symbol::All);
+    if (symbol) {
+      s = sptr;
+      return symbol->value;
+    }
+  }
+  // Final try: Another type of value.
+  {
     char *end;
-    if (*s == '$')
-      s++;
     v1 = strtol(s,&end,16);
     if (errno == ERANGE)
       throw NumericException(this,"Error: %s is out of range\n",s);
@@ -359,14 +683,6 @@ Monitor::HistoryLine::HistoryLine(const char *line)
 }
 ///
 
-/// Monitor::HistoryLine::~HistoryLine
-Monitor::HistoryLine::~HistoryLine(void)
-{
-  delete[] Line;
-  Remove();
-}
-///
-
 /// Monitor::History::History
 // Constructor of the history: Just constructs the
 // list.
@@ -376,20 +692,6 @@ Monitor::History::History(void)
 }
 ///
 
-/// Monitor::History::~History
-// Dispose the history again. Clears the list of lines
-// cached here, obviously.
-Monitor::History::~History(void)
-{    
-  struct HistoryLine *hl;
-
-  while((hl = First())) {
-    // History lines remove themselves from the history.
-    delete hl;
-  }
-}
-///
-
 /// Monitor::History::AddLine
 // Attach a new line to the history, possibly reducing
 // its size.
@@ -744,6 +1046,22 @@ void Monitor::VPrint(const char *fmt,va_list args) const
 }
 ///
 
+/// Monitor::WaitKey
+// Wait for the user to print any key to continue.
+void Monitor::WaitKey(void)
+{
+  Print("<Press RETURN to continue>\n");
+#ifdef USE_CURSES
+  int c;
+  do {
+    c = getch();
+  } while(c != KEY_ENTER && c != 0x0a && c != 0x0d && c != ' ');
+#else
+  fgetc(stdin);
+#endif
+}
+///
+
 /// Monitor::PrintCPUStatus
 // Print the CPU registers in human-readable form
 // this does less than a CPU->DisplayStatus
@@ -799,7 +1117,7 @@ bool Monitor::Command::GetAddress(UWORD &adr)
 {
   bool valid = false;
   char *token;
-  LONG value;
+  LONG value = here;
 
   token = NextToken();
   if (token) {
@@ -828,7 +1146,7 @@ bool Monitor::Command::GetDefault(int &setting,int def,int min,int max)
 	valid   = true;
 	setting = value;
       } else {
-	Print(LD " is out of range, must be >= %d and <= %d.\n",value,min,max);
+	Print(ATARIPP_LD " is out of range, must be >= %d and <= %d.\n",value,min,max);
       }
     } else {
       Print("%s is not a valid number.\n",token);
@@ -948,7 +1266,8 @@ int Monitor::Command::ReadDataLine(UBYTE *buffer,const char *prompt,char mode,bo
     return count;
   case 'D':
     base = 10;
-    // runs into the following. Base defaults to 16 otherwise.
+    // Base defaults to 16 otherwise.
+    // Intentionally falls through.
   case 'X':
     while(*input) {
       LONG value;
@@ -960,9 +1279,9 @@ int Monitor::Command::ReadDataLine(UBYTE *buffer,const char *prompt,char mode,bo
 	  count++;
 	} else {
 	  if (base == 10) {
-	    Print("Input " LD " is not a valid byte.\n",value);
+	    Print("Input " ATARIPP_LD " is not a valid byte.\n",value);
 	  } else {
-	    Print("Input " LX " is not a valid byte.\n",(ULONG)(value));
+	    Print("Input " ATARIPP_LX " is not a valid byte.\n",(ULONG)(value));
 	  }
 	  return 0;
 	}
@@ -991,14 +1310,6 @@ Monitor::Splt::Splt(class Monitor *mon,const char *lng,const char *shr,const cha
 { }
 ///
 
-/// Monitor::Splt::~Splt
-Monitor::Splt::~Splt(void)
-{
-  delete[] splitbuffer;
-  delete[] splittmp;
-}
-///
-
 /// Monitor::Splt::Apply
 // record a command to split off on top of the screen
 void Monitor::Splt::Apply(char e)
@@ -1145,7 +1456,7 @@ void Monitor::Eval::Apply(char e)
     if (token) {
       valid = monitor->EvaluateExpression(token,value);
       if (valid) {
-	Print("%s = 0x" LX " = " LD "\n",token,value,value);
+	Print("%s = 0x" ATARIPP_LX " = " ATARIPP_LD "\n",token,value,value);
       }
       return;
     }
@@ -1237,21 +1548,26 @@ ADR Monitor::UnAs::DisassembleLine(class AdrSpace *adr,ADR where,char *line)
 {
   int op;
   ADR pc = where;
+  class CPU *cpu = monitor->cpu;
   const char *name;
   UBYTE inst;
-  char buf[16];
+  char buf[48];
+  char pcbuf[48];
   Instruction::OperandType type;
+  const struct Symbol *pctarget;
+  const struct Symbol *target;
   
   if (where >= 0x10000) {
     *line = 0;
     return where;
   }
   // Format of the output:
-  // xxxx  ab cd ef  IIII noooo
-  // 0123456789abcdef0
+  // xxxx             ab cd ef  IIII noooo
+  // namenamenamenam:                namenamenamenamn,namenamenamenam    
+  // 0123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef
   inst    = adr->ReadByte(where++); 
   {
-    const struct Instruction &dis = monitor->cpu->Disassemble(inst);
+    const struct Instruction &dis = cpu->Disassemble(inst);
     name    = dis.Name;
     type    = dis.AddressingMode;
   }
@@ -1261,8 +1577,13 @@ ADR Monitor::UnAs::DisassembleLine(class AdrSpace *adr,ADR where,char *line)
     break;
   case Instruction::Immediate:
     if (where < 0x10000) {
-      op  = adr->ReadByte(where++);
-      sprintf(buf,"%-4s #$%02x",name,op);
+      op     = adr->ReadByte(where++);
+      target = Monitor::Symbol::FindSymbol(monitor->symboltable,op,Symbol::Equate,Symbol::ZeroPage);
+      if (target) {
+	sprintf(buf,"%-4s #%.16s",name,target->name);
+      } else {
+	sprintf(buf,"%-4s #$%02x",name,op);
+      }
     } else {
       sprintf(buf,"%-4s #$XX",name);
     }
@@ -1272,8 +1593,13 @@ ADR Monitor::UnAs::DisassembleLine(class AdrSpace *adr,ADR where,char *line)
     break;
   case Instruction::ZPage:
     if (where < 0x10000) {
-      op  = adr->ReadByte(where++);
-      sprintf(buf,"%-4s  $%02x",name,op);
+      op     = adr->ReadByte(where++);
+      target = Monitor::Symbol::FindSymbol(monitor->symboltable,op,Symbol::PreferLabel,Symbol::ZeroPage);
+      if (target) {
+	sprintf(buf,"%-4s  %.16s",name,target->name);
+      } else {
+	sprintf(buf,"%-4s  $%02x",name,op);
+      }
     } else {
       sprintf(buf,"%-4s  $XX",name);
     }
@@ -1281,15 +1607,25 @@ ADR Monitor::UnAs::DisassembleLine(class AdrSpace *adr,ADR where,char *line)
   case Instruction::ZPage_X:
     if (where < 0x10000) {
       op  = adr->ReadByte(where++);
-      sprintf(buf,"%-4s  $%02x,X",name,op);
+      target = Monitor::Symbol::FindSymbol(monitor->symboltable,UBYTE(op + cpu->X()),Symbol::PreferLabel,Symbol::ZeroPage);
+      if (target) {
+	sprintf(buf,"%-4s  %.16s,X",name,target->name);
+      } else {
+	sprintf(buf,"%-4s  $%02x,X",name,op);
+      }
     } else {
       sprintf(buf,"%-4s  $XX,X",name);
     }
     break;
   case Instruction::ZPage_Y:
     if (where < 0x10000) {
-      op  = adr->ReadByte(where++);
-      sprintf(buf,"%-4s  $%02x,Y",name,op);
+      op  = adr->ReadByte(where++); 
+      target = Monitor::Symbol::FindSymbol(monitor->symboltable,UBYTE(op + cpu->Y()),Symbol::PreferLabel,Symbol::ZeroPage);
+      if (target) {
+	sprintf(buf,"%-4s  %.16s,Y",name,target->name);
+      } else {
+	sprintf(buf,"%-4s  $%02x,Y",name,op);
+      }
     } else {
       sprintf(buf,"%-4s  $XX,Y",name);
     }
@@ -1297,7 +1633,12 @@ ADR Monitor::UnAs::DisassembleLine(class AdrSpace *adr,ADR where,char *line)
   case Instruction::Indirect:
     if (where < 0x10000) {
       op  = adr->ReadWord(where);
-      sprintf(buf,"%-4s ($%04x)",name,op);
+      target = Monitor::Symbol::FindSymbol(monitor->symboltable,op,Symbol::PreferEquate,Symbol::PreferAbsolute);
+      if (target) {
+	sprintf(buf,"%-4s  (%.16s)",name,target->name);
+      } else {
+	sprintf(buf,"%-4s  ($%04x)",name,op);
+      }
       where += 2;
     } else {
       sprintf(buf,"%-4s ($XXXX)",name);
@@ -1306,7 +1647,12 @@ ADR Monitor::UnAs::DisassembleLine(class AdrSpace *adr,ADR where,char *line)
   case Instruction::Indirect_X:
     if (where < 0x10000) {
       op  = adr->ReadByte(where++);
-      sprintf(buf,"%-4s ($%02x,X)",name,op);
+      target = Monitor::Symbol::FindSymbol(monitor->symboltable,UBYTE(op + cpu->X()),Symbol::PreferLabel,Symbol::ZeroPage);
+      if (target) {
+	sprintf(buf,"%-4s  (%.16s,X)",name,target->name);
+      } else {
+	sprintf(buf,"%-4s  ($%02x,X)",name,op);
+      }
     } else {
       sprintf(buf,"%-4s ($XX,X)",name);
     }
@@ -1314,7 +1660,12 @@ ADR Monitor::UnAs::DisassembleLine(class AdrSpace *adr,ADR where,char *line)
   case Instruction::Indirect_Y:
     if (where < 0x10000) {
       op  = adr->ReadByte(where++);
-      sprintf(buf,"%-4s ($%02x),Y",name,op);
+      target = Monitor::Symbol::FindSymbol(monitor->symboltable,op,Symbol::PreferLabel,Symbol::ZeroPage);
+      if (target) {
+	sprintf(buf,"%-4s  (%.16s),Y",name,target->name);
+      } else {
+	sprintf(buf,"%-4s  ($%02x),Y",name,op);
+      }
     } else {
       sprintf(buf,"%-4s ($XX),Y",name);
     }
@@ -1322,7 +1673,12 @@ ADR Monitor::UnAs::DisassembleLine(class AdrSpace *adr,ADR where,char *line)
   case Instruction::Indirect_Z:
     if (where < 0x10000) {
       op  = adr->ReadByte(where++);
-      sprintf(buf,"%-4s ($%02x)",name,op);
+      target = Monitor::Symbol::FindSymbol(monitor->symboltable,op,Symbol::PreferLabel,Symbol::ZeroPage);
+      if (target) {
+	sprintf(buf,"%-4s  (%.16s)",name,target->name);
+      } else {
+	sprintf(buf,"%-4s  ($%02x)",name,op);
+      }
     } else {
       sprintf(buf,"%-4s ($XX)",name);
     }
@@ -1330,7 +1686,12 @@ ADR Monitor::UnAs::DisassembleLine(class AdrSpace *adr,ADR where,char *line)
   case Instruction::Absolute:
     if (where < 0xffff) {
       op  = adr->ReadWord(where);
-      sprintf(buf,"%-4s  $%04x",name,op);
+      target = Monitor::Symbol::FindSymbol(monitor->symboltable,op,Symbol::PreferLabel,Symbol::PreferAbsolute);
+      if (target) {
+	sprintf(buf,"%-4s  %.16s",name,target->name);
+      } else {
+	sprintf(buf,"%-4s  $%04x",name,op);
+      }
       where += 2;
     } else {
       sprintf(buf,"%-4s  $XXXX",name);
@@ -1339,7 +1700,12 @@ ADR Monitor::UnAs::DisassembleLine(class AdrSpace *adr,ADR where,char *line)
   case Instruction::Absolute_X:
     if (where < 0xffff) {
       op  = adr->ReadWord(where);
-      sprintf(buf,"%-4s  $%04x,X",name,op);
+      target = Monitor::Symbol::FindSymbol(monitor->symboltable,UWORD(op + cpu->X()),Symbol::PreferLabel,Symbol::PreferAbsolute);
+      if (target) {
+	sprintf(buf,"%-4s  %.16s,X",name,target->name);
+      } else {
+	sprintf(buf,"%-4s  $%04x,X",name,op);
+      }
       where += 2;
     } else {
       sprintf(buf,"%-4s  $XXXX,X",name);
@@ -1348,7 +1714,12 @@ ADR Monitor::UnAs::DisassembleLine(class AdrSpace *adr,ADR where,char *line)
   case Instruction::Absolute_Y:
     if (where < 0xffff) {
       op  = adr->ReadWord(where);
-      sprintf(buf,"%-4s  $%04x,Y",name,op);
+      target = Monitor::Symbol::FindSymbol(monitor->symboltable,UWORD(op + cpu->Y()),Symbol::PreferLabel,Symbol::PreferAbsolute);
+      if (target) {
+	sprintf(buf,"%-4s  %.16s,Y",name,target->name);
+      } else {
+	sprintf(buf,"%-4s  $%04x,Y",name,op);
+      }
       where += 2;
     } else {
       sprintf(buf,"%-4s  $XXXX,Y",name);
@@ -1357,7 +1728,12 @@ ADR Monitor::UnAs::DisassembleLine(class AdrSpace *adr,ADR where,char *line)
   case Instruction::AbsIndirect_X:
     if (where < 0xffff) {
       op  = adr->ReadWord(where);
-      sprintf(buf,"%-4s ($%04x,X)",name,op);
+      target = Monitor::Symbol::FindSymbol(monitor->symboltable,UBYTE(op + cpu->X()),Symbol::PreferLabel,Symbol::PreferAbsolute);
+      if (target) {
+	sprintf(buf,"%-4s  (%.16s,X)",name,target->name);
+      } else {
+	sprintf(buf,"%-4s  ($%04x,X)",name,op);
+      }
       where += 2;
     } else {
       sprintf(buf,"%-4s ($XXXX,X)",name);
@@ -1367,7 +1743,12 @@ ADR Monitor::UnAs::DisassembleLine(class AdrSpace *adr,ADR where,char *line)
     // Note that this displacement is signed
     if (where < 0x10000) {
       op  = (BYTE)(adr->ReadByte(where++));
-      sprintf(buf,"%-4s  $%04x",name,where+op);
+      target = Monitor::Symbol::FindSymbol(monitor->symboltable,where+op,Symbol::PreferLabel,Symbol::PreferAbsolute);
+      if (target) {
+	sprintf(buf,"%-4s  %.16s",name,target->name);
+      } else {
+	sprintf(buf,"%-4s  $%04x",name,where+op);
+      }
     } else {
       sprintf(buf,"%-4s  $XXXX",name);
     }
@@ -1376,7 +1757,21 @@ ADR Monitor::UnAs::DisassembleLine(class AdrSpace *adr,ADR where,char *line)
     if (where < 0xffff) {
       UBYTE zp = adr->ReadByte(where++);
       op       = (BYTE)adr->ReadByte(where++);
-      sprintf(buf,"%-4s  $%02x,$%04x",name,zp,where + op);
+      pctarget = Monitor::Symbol::FindSymbol(monitor->symboltable,where+op,Symbol::PreferLabel,Symbol::PreferAbsolute);
+      target   = Monitor::Symbol::FindSymbol(monitor->symboltable,zp      ,Symbol::PreferLabel,Symbol::ZeroPage);
+      if (target) {
+	if (pctarget) {
+	  sprintf(buf,"%-4s  %.16s,%.16s",name,target->name,pctarget->name);
+	} else {
+	  sprintf(buf,"%-4s  %.16s,$%04x",name,target->name,where+op);
+	}
+      } else {
+	if (pctarget) {
+	  sprintf(buf,"%-4s  $%02x,%.16s",name,zp,pctarget->name);
+	} else {
+	  sprintf(buf,"%-4s  $%02x,$%04x",name,zp,where+op);
+	}
+      }
     } else {
       sprintf(buf,"%-4s  $XX,$XXXX"  ,name);
     }
@@ -1385,15 +1780,21 @@ ADR Monitor::UnAs::DisassembleLine(class AdrSpace *adr,ADR where,char *line)
   //
   // Now disassemble the remaining line dependent on the number of arguments
   // we pulled off (one to three).
+  pctarget = Monitor::Symbol::FindSymbol(monitor->symboltable,pc,Symbol::Any,Symbol::Absolute);
+  if (pctarget) {
+    sprintf(pcbuf,"$%04x:%.16s",pc,pctarget->name);
+  } else {
+    sprintf(pcbuf,"$%04x:",pc);
+  }
   switch(where - pc) {
   case 1:
-    sprintf(line,"$%04x: %02x        %s",pc,inst,buf);
+    sprintf(line,"%-22s %02x        %s",pcbuf,inst,buf);
     break;
   case 2:
-    sprintf(line,"$%04x: %02x %02x     %s",pc,inst,adr->ReadByte(pc+1),buf);
+    sprintf(line,"%-22s %02x %02x     %s",pcbuf,inst,adr->ReadByte(pc+1),buf);
     break;
   case 3:
-    sprintf(line,"$%04x: %02x %02x %02x  %s",pc,inst,adr->ReadByte(pc+1),
+    sprintf(line,"%-22s %02x %02x %02x  %s",pcbuf,inst,adr->ReadByte(pc+1),
 	    adr->ReadByte(pc+2),buf);
     break;
   }
@@ -1401,11 +1802,11 @@ ADR Monitor::UnAs::DisassembleLine(class AdrSpace *adr,ADR where,char *line)
   // Check whether we are at the CPU PC position. If so, print
   // a star behind the address.
   if (pc == monitor->cpu->PC()) {
-    line[16] = '*';
+    line[32] = '*';
   }
   // Check whether we have a break point at this position.
   if (monitor->cpu->IfBreakPoint(pc)) {
-    line[15] = 'B';
+    line[31] = 'B';
   }
 
   return where;
@@ -1508,13 +1909,6 @@ Monitor::Step::Step(class Monitor *mon,const char *lng,const char *shr,const cha
 { }
 ///
 
-/// Monitor::Step::~Step
-Monitor::Step::~Step(void)
-{
-  delete[] lineaddresses;
-}
-///
-
 /// Monitor::Step::OpenDisplay
 // Create the optimized tracder display
 bool Monitor::Step::OpenDisplay(void)
@@ -1614,7 +2008,7 @@ bool Monitor::Step::RefreshLine(class AdrSpace *adr,ADR pc,bool showpc)
       if (lineaddresses[lines] == pc) {
 	monitor->UnAs.DisassembleLine(adr,pc,buffer);
 	if (!showpc)
-	  buffer[16] = ' ';
+	  buffer[32] = ' ';
 	move(y,0);
 	deleteln();
 	insertln();
@@ -1697,6 +2091,8 @@ bool Monitor::Step::MainLoop(void)
     case KEY_F(10):
     case 'n':
     case 'N':
+    case 't':
+    case 'T':
       // Next command.
       // Check the CPU command at the current address. If this is a branch instruction,
       // we continue until we arrive at a PC larger than the current address. Otherwise,
@@ -1927,7 +2323,8 @@ void Monitor::Stat::Apply(char e)
       Print("Unknown emulator component %s\n",token);
       return;
     }
-    MissingArg();  
+    MissingArg();
+    break;
   default:
     ExtInvalid();   
   }
@@ -1967,7 +2364,7 @@ void Monitor::Edit::Apply(char e)
       do {
 	UBYTE *p;
 	int   count;
-	char  adbuffer[8];
+	char  adbuffer[9];
 	UBYTE buffer[128];
 	//
 	sprintf(adbuffer,"$%04x : ",here);
@@ -2232,36 +2629,36 @@ void Monitor::SetR::Apply(char e)
     value     = 0;
     if (monitor->EvaluateExpression(valstr,value)) {
       if (value<0x0000 || value > 0xffff) {
-	Print("Register value " LX " out of range\n",(ULONG)value);
+	Print("Register value " ATARIPP_LX " out of range\n",(ULONG)value);
 	return;
       }
       if (!strcasecmp(setstr,"A")) {
 	if (value > 0xff) {
-	  Print("Register value " LX " out of range\n",(ULONG)value);
+	  Print("Register value " ATARIPP_LX " out of range\n",(ULONG)value);
 	  return;
 	}
 	monitor->cpu->A() = UBYTE(value);
       } else if (!strcasecmp(setstr,"X")) {
 	if (value > 0xff) {
-	  Print("Register value " LX " out of range\n",(ULONG)value);
+	  Print("Register value " ATARIPP_LX " out of range\n",(ULONG)value);
 	  return;
 	}
 	monitor->cpu->X() = UBYTE(value);
       } else if (!strcasecmp(setstr,"Y")) {
 	if (value > 0xff) {
-	  Print("Register value " LX " out of range\n",(ULONG)value);
+	  Print("Register value " ATARIPP_LX " out of range\n",(ULONG)value);
 	  return;
 	}
 	monitor->cpu->Y() = UBYTE(value);
       } else if (!strcasecmp(setstr,"S")) {
 	if (value > 0xff) {
-	  Print("Register value " LX " out of range\n",(ULONG)value);
+	  Print("Register value " ATARIPP_LX " out of range\n",(ULONG)value);
 	  return;
 	}
 	monitor->cpu->S() = UBYTE(value);
       } else if (!strcasecmp(setstr,"P")) {
 	if (value > 0xff) {
-	  Print("Register value " LX " out of range\n",(ULONG)value);
+	  Print("Register value " ATARIPP_LX " out of range\n",(ULONG)value);
 	  return;
 	}
 	monitor->cpu->P() = UBYTE(value);
@@ -2328,7 +2725,7 @@ ADR Monitor::Dlst::DisassembleLine(class AdrSpace *adr,ADR where,char *line)
   bool load    = false;
   bool waitvbr = false;
   char cmdname[20],prehex[20];
-  char options[40];
+  char options[33];
 
   inst    = adr->ReadByte(where);
   if (inst & 0x80) {
@@ -2399,10 +2796,10 @@ ADR Monitor::Dlst::DisassembleLine(class AdrSpace *adr,ADR where,char *line)
     strcat(options," DLI");
   
   if (load) {
-    sprintf(line,"%s %s @$%04x %s",prehex,cmdname,adr->ReadWord(where+1),options);
+    snprintf(line,80,"%s %s @$%04x %s",prehex,cmdname,adr->ReadWord(where+1),options);
     where += 3;
   } else {
-    sprintf(line,"%s %s %s",prehex,cmdname,options);
+    snprintf(line,80,"%s %s %s",prehex,cmdname,options);
     where++;
   }
 
@@ -2464,6 +2861,8 @@ void Monitor::Envi::Apply(char e)
   case '?':
     Print("ENVI.A : toggle between CPU and ANTIC address space\n");
     Print("ENVI.L [filename] : set tracing output file\n");
+    Print("ENVI.S [filename] : load ld65 debug symbols from file\n");
+    Print("ENVI.C : clear symbol table\n");
     break;
   case 'A':
     if (monitor->currentadr == monitor->cpuspace) {
@@ -2474,6 +2873,20 @@ void Monitor::Envi::Apply(char e)
       Print("Current address space is CPU.\n");
     }
     break;
+  case 'C':
+    monitor->ClearSymbolTable();
+    Print("Symbol table removed.\n");
+    break;
+  case 'S':
+    token = NextToken();
+    if (token) {
+      if (monitor->ParseSymbolTable(token)) {
+	Print("Symbols from %s added to the symbol table.\n",token);
+      } else {
+	Print("No symbols found in %s.\n",token);
+      }
+    }
+    break;
   case 'L':
     if (monitor->tracefile) {
       fclose(monitor->tracefile);
@@ -2544,7 +2957,8 @@ void Monitor::BrkP::Apply(char e)
   switch(e) {
   case '?':
     Print("BRKP.S [addr] : set breakpoint at address\n"
-	  "BRKP.W [addr] : set watchpoint at address\n"
+	  "BRKP.W [addr] : set write only watchpoint at address\n"
+	  "BRKP.V [addr] : set read/write watchpoint at address\n"
 	  "BRKP.C [addr] : clear breakpoint at address\n"
 	  "BRKP.D [addr] : disable breakpoint at address\n"
 	  "BRKP.E [addr] : enable breakpoint at address\n"
@@ -2577,6 +2991,7 @@ void Monitor::BrkP::Apply(char e)
     }
     break;  
   case 'W':
+  case 'V':
     if (GetAddress(here)) {
       if (LastArg()) {
 	for(i=0;i<NumBrk;i++) {
@@ -2587,11 +3002,12 @@ void Monitor::BrkP::Apply(char e)
 	}
 	for(i=0;i<NumBrk;i++) {
 	  if (WatchPoints[i].id == -1) {
-	    WatchPoints[i].id = MMU->DebugRAM()->SetWatchPoint(here);
+	    WatchPoints[i].id = MMU->DebugRAM()->SetWatchPoint(here,e == 'V');
 	    if (WatchPoints[i].id >= 0) {
 	      cpu->EnableWatchPoints();
 	      WatchPoints[i].address = here;
 	      WatchPoints[i].enabled = true;
+	      WatchPoints[i].read    = (e == 'V');
 	      Print("Installed watchpoint at address : $%04x\n",here);
 	      return;
 	    }
@@ -2665,7 +3081,7 @@ void Monitor::BrkP::Apply(char e)
 	}	
 	for(i=0;i<NumBrk;i++) {
 	  if (WatchPoints[i].id >= 0 && WatchPoints[i].address == here) {
-	    WatchPoints[i].id = MMU->DebugRAM()->SetWatchPoint(here);
+	    WatchPoints[i].id = MMU->DebugRAM()->SetWatchPoint(here,WatchPoints[i].read);
 	    if (WatchPoints[i].id >= 0) {
 	      cpu->EnableWatchPoints();
 	      WatchPoints[i].address = here;
@@ -2711,7 +3127,8 @@ void Monitor::BrkP::Apply(char e)
       }
       for(i=0;i<NumBrk;i++) {
 	if (WatchPoints[i].id >= 0) {
-	  Print("Watchpoint at $%04x (%s)\n",
+	  Print("Watchpoint(%s) at $%04x (%s)\n",
+		(WatchPoints[i].read)?("read/write"):("write only"),
 		WatchPoints[i].address,
 		(WatchPoints[i].enabled)?("enabled"):("disabled"));
 	}
@@ -2908,6 +3325,158 @@ void Monitor::Disk::Apply(char e)
 }
 ///
 
+/// Monitor::Prof::Prof
+// Build the profiler.
+Monitor::Prof::Prof(class Monitor *mon,const char *lng,const char *shr,const char *helper)
+  : Command(mon,lng,shr,helper,'L')
+{ }
+///
+
+/// Monitor::Prof::Apply
+// The monitor implementation.
+void Monitor::Prof::Apply(char e)
+{
+  switch(e) {
+  case '?':
+    Print("Profiler subcommands:\n"
+	  "PROF.S : start profiling\n"
+	  "PROF.X : stop profiling\n"
+	  "PROF.L : list profile data\n"
+	  "PROF.C : list cumulative profiling data\n");
+    break;
+  case 'S':
+    if (monitor->cpu->ProfilingCountersOf()) {
+      Print("Profiler is already running.\n");
+    } else {
+      monitor->cpu->StartProfiling();
+      Print("Profiling enabled.\n");
+    }
+    break;
+  case 'X':
+    if (monitor->cpu->ProfilingCountersOf()) {
+      monitor->cpu->StopProfiling();
+      Print("Profiler stopped.\n");
+    } else {
+      Print("Profiler is not running.\n");
+    }
+    break;
+  case 'L':
+  case 'C':
+    {
+      struct Entry {
+	struct Entry *next;
+	ULONG count;
+	ADR   pc;
+      } *entries         = NULL,*last = NULL;
+      const ULONG *cntrs = (e == 'L')?
+	(monitor->cpu->ProfilingCountersOf()):
+	(monitor->cpu->CumulativeProfilingCountersOf());
+      int lines;
+      int height  = 32;
+      ADR pc      = 0;
+      UQUAD total = 0;
+#ifdef USE_CURSES
+      WINDOW *window = (WINDOW *)monitor->curses->window;
+      getmaxyx(window,height,lines);
+#endif
+      if (cntrs == NULL) {
+	Print("Profiler is currently not running. Please start the profiler first with\n"
+	      "PROF.S, run the program, then use PROF.L again to show collected data.\n");
+	break;
+      }
+      try {
+	do {
+	  if (cntrs[pc]) {
+	    // Ok, so this address was used after all. Check whether its count is the
+	    // same as that of the last entry (i.e. the PC above). If so, do not collect
+	    // because it is likely the same function.
+	    if (last == NULL || last->count != cntrs[pc]) {
+	      struct Entry **i;
+	      struct Entry *e = new struct Entry;
+	      e->count = cntrs[pc];
+	      e->pc    = pc;
+	      total   += cntrs[pc];
+	      // Run a stupid insertion sort to sort entries by size...
+	      i = &entries;
+	      while(*i && (*i)->count >= e->count)
+		i = &((*i)->next);
+	      // insert before i
+	      e->next = *i;
+	      *i      = e;
+	      last    = e;
+	    }
+	  }
+	  pc++;
+	} while(pc != 0xffff);
+	if (e == 'C')
+	  total = cntrs[pc];
+	//
+	// Now print the profiling entries.
+	last  = entries;
+	lines = 0;
+	while(last) {
+	  const struct Symbol *target;
+	  target = Monitor::Symbol::FindSymbol(monitor->symboltable,last->pc,Symbol::Label,Symbol::PreferAbsolute);
+	  //
+	  // Print the label name if we have it, or the label.
+	  if (target) {
+	    Print("%-22s %10lu (%.3f%%)\n",target->name,(unsigned long)(last->count),
+		  100.0*last->count/double(total));
+	  } else {
+	    Print("%4x                   %10lu (%.3f%%)\n",(unsigned int)last->pc,
+		  (unsigned long)(last->count),
+		  100.0*last->count/double(total));
+	  }
+	  lines++;
+	  if (lines >= (height >> 1)) {
+	    const char *prompt = "*** Press RETURN to continue or Q to abort ***";
+	    int in = 0;
+#ifndef USE_CURSES
+	    char input[64];
+	    
+	    printf("%s",prompt);
+	    fflush(stdout);
+	    monitor->machine->RefreshDisplay();
+	    fgets(input,sizeof(input),stdin);
+	    in = input[0];
+#else
+	    monitor->Print("%s",prompt);
+	    do {
+	      in = getch();
+	    } while(in == ERR);
+	    monitor->Print("\n");
+#endif	   
+	    if (in == 'q' || in == 'Q')
+	      break;
+	    lines = 0;
+#ifdef USE_CURSES
+	    {
+	      int x,y;
+	      getyx(window,y,x);
+	      wmove(window,y-1,0);
+	      wdeleteln(window);
+	    }
+#endif
+	  }
+	  last = last->next;
+	}
+	while((last = entries)) {
+	  entries = entries->next;
+	  delete last;
+	}
+      } catch(...) {
+	while((last = entries)) {
+	  entries = entries->next;
+	  delete last;
+	}
+	throw;
+      }
+    }
+    break;
+  }
+}
+///
+
 /// Monitor::ParseCmd
 // Get the current token from the command line, interpret as an argument
 // and run the corresponding entry
@@ -3043,6 +3612,12 @@ void Monitor::UnknownESC(UBYTE code)
 	"emulator with the RSET command.\n",
 	code,cpu->PC()-2); // the CPU advanced the PC already
   MainLoop();
+#ifdef MUST_OPEN_CONSOLE
+  if (!fetchtrace) {
+	machine->Display()->SwitchScreen(true);
+	CloseConsole();
+  }
+#endif
 }
 ///
 
@@ -3063,6 +3638,12 @@ void Monitor::Crash(UBYTE code)
 	"emulator with the RSET command.\n",
 	code,cpu->PC());
   MainLoop();
+#ifdef MUST_OPEN_CONSOLE
+  if (!fetchtrace) {
+	machine->Display()->SwitchScreen(true);
+	CloseConsole();
+  }
+#endif
 }
 ///
 
@@ -3083,6 +3664,12 @@ void Monitor::Jam(UBYTE code)
 	"emulator with the RSET command.\n",
 	code,cpu->PC());
   MainLoop();
+#ifdef MUST_OPEN_CONSOLE
+  if (!fetchtrace) {
+	machine->Display()->SwitchScreen(true);
+	CloseConsole();
+  }
+#endif
 }
 ///
 
@@ -3099,6 +3686,12 @@ void Monitor::EnterMonitor(void)
   Step.CloseDisplay();
   Print("\nEntering monitor\n");
   MainLoop();
+#ifdef MUST_OPEN_CONSOLE
+  if (!fetchtrace) {
+	machine->Display()->SwitchScreen(true);
+	CloseConsole();
+  }
+#endif
 }
 ///
 
@@ -3122,6 +3715,12 @@ void Monitor::CapturedBreakPoint(int,ADR pc)
     MainLoop();
   }
   curses = NULL;
+#ifdef MUST_OPEN_CONSOLE
+  if (!fetchtrace) {
+	machine->Display()->SwitchScreen(true);
+	CloseConsole();
+  }
+#endif
 }
 ///
 
@@ -3145,6 +3744,12 @@ void Monitor::CapturedWatchPoint(int,ADR mem)
     MainLoop();
   }
   curses = NULL;
+#ifdef MUST_OPEN_CONSOLE
+  if (!fetchtrace) {
+	machine->Display()->SwitchScreen(true);
+	CloseConsole();
+  }
+#endif
 }
 ///
 
@@ -3192,6 +3797,12 @@ void Monitor::CapturedTrace(ADR)
       MainLoop(false);
     }
     curses = NULL;
+#ifdef MUST_OPEN_CONSOLE
+	if (!fetchtrace) {
+		machine->Display()->SwitchScreen(true);
+		CloseConsole();
+	}
+#endif
   }
 }
 ///
