@@ -2,7 +2,7 @@
  **
  ** Atari++ emulator (c) 2002 THOR-Software, Thomas Richter
  **
- ** $Id: osrom.cpp,v 1.48 2010-11-06 16:43:25 thor Exp $
+ ** $Id: osrom.cpp,v 1.54 2021/08/16 10:31:01 thor Exp $
  **
  ** In this module: Administration/loading of Os ROMs
  **********************************************************************************/
@@ -61,6 +61,8 @@ OsROM::OsROM(class Machine *mach)
 /// OsROM::~OsROM
 OsROM::~OsROM(void)
 {
+  PatchProvider::DisposePatches();
+  
   delete[] osapath;
   delete[] osbpath;
   delete[] os1200path;
@@ -149,8 +151,8 @@ void OsROM::PatchFromDump(const unsigned char *dump,int pages)
 // Erase the math pack, replace by HLT instructions.
 void OsROM::KillMathPack(void)
 {
-  class RomPage *page = rom + ((0xd800 - 0xc000) >> PAGE_SHIFT);
-  int pages           = (0xe000 - 0xd800) >> PAGE_SHIFT;
+  class RomPage *page = rom + ((0xd800 - 0xc000) >> Page::Page_Shift);
+  int pages           = (0xe000 - 0xd800) >> Page::Page_Shift;
   UBYTE offset;
   //
   do {
@@ -443,7 +445,7 @@ void OsROM::ParseArgs(class ArgParser *args)
     args->DefineBool("InstallEDevice","install E: CIO patch",epatch);
     args->DefineBool("InstallRDevice","install R: CIO patch",rpatch);
     args->DefineBool("InstallHAsDisk","install host handler as D: device",hAsd); 
-#if HAVE_MATHPACKPATCH   
+#if HAVE_MATH
     args->DefineBool("InstallMathPatch","install fast math pack patch",mppatch);
 #endif
 
@@ -562,10 +564,10 @@ void OsROM::DisplayStatus(class Monitor *mon)
 		   "\tOsBPath       : %s\n"
 		   "\tOs1200Path    : %s\n"
 		   "\tOsXLPath      : %s\n"
-		   "\tH1 diretory   : %s\n"
-		   "\tH2 diretory   : %s\n"
-		   "\tH3 diretory   : %s\n"
-		   "\tH4 diretory   : %s\n",
+		   "\tH1 directory  : %s\n"
+		   "\tH2 directory  : %s\n"
+		   "\tH3 directory  : %s\n"
+		   "\tH4 directory  : %s\n",
 		   (siopatch)?("on"):("off"),
 		   (ppatch)?("on"):("off"),
 		   (hpatch)?("on"):("off"),
