@@ -2,7 +2,7 @@
  **
  ** Atari++ emulator (c) 2002 THOR-Software, Thomas Richter
  **
- ** $Id: pokey.cpp,v 1.121 2013-01-01 17:36:55 thor Exp $
+ ** $Id: pokey.cpp,v 1.144 2022/12/20 18:01:33 thor Exp $
  **
  ** In this module: Pokey emulation 
  **
@@ -24,6 +24,7 @@
 #include "monitor.hpp"
 #include "gamecontroller.hpp"
 #include "pokey.hpp"
+#include "tape.hpp"
 #include "cpu.hpp"
 #include "snapshot.hpp"
 #include "time.hpp"
@@ -60,11 +61,12 @@ const UBYTE Pokey::PolyCounter5[Poly5Size] =
 // Pokey constructor
 Pokey::Pokey(class Machine *mach,int unit)
   : Chip(mach,(unit)?"ExtraPokey":"Pokey"), Saveable(mach,(unit)?"ExtraPokey":"Pokey"), 
-    HBIAction(mach), CycleAction(mach), IRQSource(mach),
+    VBIAction(mach), HBIAction(mach), CycleAction(mach), IRQSource(mach),
     Outcnt(0), 
     PolyCounter9(new UBYTE[Poly9Size]), PolyCounter17(new UBYTE[Poly17Size]),
     Random9(new UBYTE[Poly9Size]), Random17(new UBYTE[Poly17Size]),
-    SampleCnt(0), OutputMapping(new BYTE[256]), Unit(unit)
+    SampleCnt(0), OutputMapping(new BYTE[256]), Unit(unit), 
+    SAPOutput(NULL), SongName(NULL), AuthorName(NULL), EnableSAP(false)
 {
   int n;
   //
@@ -134,17 +136,21 @@ Pokey::Pokey(class Machine *mach,int unit)
   AllPot           = 0x00;
   // PAL operation
   NTSC             = false;
+  isAuto           = true;
   SIOSound         = true;
   CycleTimers      = false;
-  LongStartBit     = false;
   Outcnt           = 0;
   //
   // Initialize serial timing, will be overridden
   // by SkCtrl writes
   SerIn_Delay      = 9 * Base15kHz;
   SerOut_Delay     = 9 * Base15kHz;
+  SerIn_Clock      = 9 * Base15kHz;
+  SerOut_Clock     = 9 * Base15kHz;
   SerXmtDone_Delay = 9 * Base15kHz;
   SerBitOut_Delay  = Base15kHz;
+  SerInRate        = 0;
+  SerInManual      = false;
   //
   // Initialize pointers to be on the safe side.
   sound            = NULL;
@@ -207,6 +213,8 @@ void Pokey::WarmStart(void)
   SkStat             = 0xf0; // no pending errors
   SkCtrl             = 0x00;
   SerInBytes         = 0x00; // reset serial port
+  SerInRate          = 0;
+  SerInManual        = false;
   //
   //
   // Initialize poly counter registers
@@ -265,6 +273,16 @@ void Pokey::ColdStart(void)
 // Destructor of the pokey class
 Pokey::~Pokey(void)
 {
+  
+  // Shut down the SAP output in case it is
+  // still running.
+  if (SAPOutput) {
+    fclose(SAPOutput);
+  }
+
+  delete[] SongName;
+  delete[] AuthorName;
+
   // If the cycle-precise timers were active, remove it here.
   if (CycleTimers) {
     Node<CycleAction>::Remove();
@@ -275,6 +293,8 @@ Pokey::~Pokey(void)
   // allocated.
   delete[] PolyCounter9;
   delete[] PolyCounter17;
+  delete[] Random9;
+  delete[] Random17;
   delete[] OutputMapping;
 }
 ///
@@ -339,7 +359,7 @@ void Pokey::UpdateSound(UBYTE mask)
 
 	if (AudioCtrl & Mhz17Flag[n]) {
 	  ch->DivNMax    = ch->AudioF + 7; // clock by 1.79Mhz
-	  ch->DivFullMax = 255        + 7; // full wraparound required.
+	  ch->DivFullMax = 255        + 1; // full wraparound required. FIX: No wraparound full reset required.
 	} else {
 	  ch->DivNMax    = (ch->AudioF + 1) * TimeBase;
 	  ch->DivFullMax = (255        + 1) * TimeBase;
@@ -369,6 +389,11 @@ void Pokey::UpdateSound(UBYTE mask)
 	// is also on. Note that we checked the filtered channel in a previous
 	// loop.
 	ch->ChannelOn   = ch[-2].ChannelOn;
+      } else if (AudioCtrl & LinkHiFlag[n]) {
+	// Also unmute if this is the high-part of a linked filter pair because
+	// the frequency of the high-filter now impacts the maximum cycle count
+	// of the low filter.
+	ch->ChannelOn   = ch[-1].ChannelOn;
       } else {
 	// Otherwise, disable the channel unless we have reason not to.
 	// We also clear the hi-flop of the filtered channel here to reset
@@ -380,7 +405,9 @@ void Pokey::UpdateSound(UBYTE mask)
       // VolOnly is turned off and the volume is larger than zero.
       // Unless the channel has volonly set or has a divisor which is too
       // low. 1.79MHz / 22kHz = 81.
-      audc = ch->AudioC;
+      audc       = ch->AudioC;
+      ch->AudioV = audc & 0x0f; // volume
+      ch->AudioP = audc >> 5;   // polycounter
       if ((audc & 0x0f) && (audc & 0x10) == 0) {
 	audc &= 0xe0;
 	if (audc != 0xa0 && audc != 0xe0) {
@@ -446,6 +473,7 @@ void Pokey::UpdateSound(UBYTE mask)
     if (mask & 0x0c) {
       // Input is driven by channel 3 (or 2 and 3)
       SerIn_Delay      = 20 * Ch[3].DivNMax;
+      SerIn_Clock      = UWORD((Ch[3].DivNMax < 0xffff)?(Ch[3].DivNMax):(0xffff));
     }
   }
   // Check the output clock mode. This is controlled
@@ -460,6 +488,7 @@ void Pokey::UpdateSound(UBYTE mask)
     // Output clocked by channel 3 (or 2 and 3)
     if (mask & 0x0c) {
       SerOut_Delay     = Ch[3].DivNMax;
+      SerOut_Clock     = UWORD((Ch[3].DivNMax < 0xffff)?(Ch[3].DivNMax):(0xffff));
       SerBitOut_Delay  = 2  * SerOut_Delay;
       SerXmtDone_Delay = 20 * SerOut_Delay;
       SerOut_Delay     = (SerOut_Delay > 10)?(SerOut_Delay - 10):(1);
@@ -515,10 +544,12 @@ void Pokey::UpdateSound(UBYTE mask)
 	    Ch[2].ChannelOn = true;
 	  if (Ch[3].AudioC & 0x0f) {
 	    Ch[3].ChannelOn = true;
-	    // FIXME: Ugly hack to get the sound 
-	    // approximately right.
-	    // See above for how it really works.
-	    Ch[3].DivNMax  *= 21;
+	    if (SerInRate < 500) {
+	      // FIXME: Ugly hack to get the sound 
+	      // approximately right.
+	      // See above for how it really works.
+	      Ch[3].DivNMax  *= 21;
+	    }
 	  }
 	} else {
 	  // Transfer blocked, stop it.
@@ -550,6 +581,46 @@ void Pokey::ComputeSamples(struct AudioBufferBase *to,int size,int dspsamplerate
   ULONG SampleMax    = (Frequency17Mhz << 8) / dspsamplerate;
   LONG  offset       = delta;
   
+  if (SIOSound) {
+    // Check whether the tape is running and some data comes actually in.
+    // If so, misuse the audio channels to re-generate the sound that would
+    // normally come from the tape.
+    if (AudioCtrl == 0x28) {
+      bool bit   = true;
+      //
+      if (SkCtrl & 0x10) {
+	if (SerIn_Counter > 0 && SerInBuffer && SerInRate > 0 && SerInBytes > 0) {
+	  int bitposition   = (SerIn_Counter + SerInRate - 1) / SerInRate; // "half-bits"
+	  //
+	  if (bitposition == 20 || bitposition == 19) { 
+	    // The start bit. Actually, 1 1/2 start bits. If we assume that the
+	    // receiver waits 1 1/2 bits to sample at the middle of the bit, this
+	    // might be ok.
+	    bit   = false;
+	  } else if (bitposition < 19 && bitposition > 2) {
+	    bit   = ((*SerInBuffer >> (7 - ((bitposition - 3) >> 1))) & 0x01)?true:false;
+	  }
+	}
+	Ch[3].AudioV = Ch[3].AudioC & 0x0f;
+	if (bit)
+	  Ch[3].AudioV = (Ch[3].AudioV * 3) >> 2;
+      }
+      if (sio && sio->isMotorEnabled()) {
+	class Tape *tape = machine->Tape();
+	if (tape && tape->isPlaying() && tape->isRecording() == false) {
+	  // Misuse channel 3 to generate the sound. It is actually the serial counter.
+	  Ch[2].ChannelOn = true;
+	  Ch[2].AudioV    = 0x08;
+	  Ch[2].AudioP    = 5;
+	  if (bit) {
+	    Ch[2].DivNMax   = (0x05 + 1) * Base64kHz;
+	  } else {
+	    Ch[2].DivNMax   = (0x07 + 1) * Base64kHz;
+	  }
+	}
+      }
+    }
+  }
 
   if ((SkCtrl & 0x03) == 0) {
     // Sound completely disabled.
@@ -602,7 +673,7 @@ void Pokey::ComputeSamples(struct AudioBufferBase *to,int size,int dspsamplerate
 	  mask = 0x00;
 	}
       }
-      current += mask & c->AudioC;
+      current += mask & c->AudioV;
     }
     //
     // Now generate the output:
@@ -667,7 +738,7 @@ void Pokey::ComputeSamples(struct AudioBufferBase *to,int size,int dspsamplerate
       // Now check the AudioControl for which polycounter we should
       // use for updating. All this information is encoded in the polypointer
       // array indexed by the audc >> 5.
-      audc = c->AudioC >> 5;
+      audc = c->AudioP;
       // Check for the first gate. This is the polycounter #5 or no polycounter
       // at all.
       if (**PolyPointerFirst[audc]) {
@@ -689,7 +760,7 @@ void Pokey::ComputeSamples(struct AudioBufferBase *to,int size,int dspsamplerate
       if (SkCtrl & 0x08) {
 	switch(next_event) {
 	case 1: // Channel 1 syncs channel 0 always.
-	  Ch[0].DivNCnt = Ch[0].DivNMax;
+	  Ch[0].DivNCnt   = Ch[0].DivNMax;
 	  break;
 	case 0: // Channel 0 syncs channel 1 if the serial register is set.
 	  if ((SerOut_Register & 0x01) && (SkCtrl & 0x80) == 0)
@@ -793,6 +864,8 @@ void Pokey::UpdatePots(int steps)
 // Forward the pokey status for N steps.
 void Pokey::GoNSteps(int steps)
 {
+  static const UBYTE irqbits[4]    = { 0x01, 0x02, 0x00, 0x04 };
+  static const UBYTE LinkLoFlag[4] = { 0x10, 0x00, 0x08, 0x00 };
   struct Channel *c;
   int ch,lastch;
   //
@@ -818,18 +891,36 @@ void Pokey::GoNSteps(int steps)
     //
     // Check for serial input done.
     if (SerIn_Counter > 0) {
+      bool update = (SerInBytes && SerIn_Counter > SerIn_Delay)?true:false;
       if ((SerIn_Counter -= steps) <= 0) {  
-	SerIn_Counter = 0;
-	// Check whether SIO delivered any new bytes. If so,
-	// generate an IRQ to deliver them.
-	if (SerInBytes) {
-	  // Serial input just finished. Generate an IRQ.   
-	  GenerateIRQ(0x20);
-	} else if (sio) {
-	  // Ask SIO again.
-	  sio->RequestInput();
+	// If we parsed the line manually, check whether
+	// the stop bit was parsed as well. If so, the data
+	// is here (incorrectly) not passed into the serial shift
+	// register but removed manually, and serial input proceeds.
+	if (SerInManual && SerInRate) {
+	  SerIn_Counter += SerInRate * 20;
+	  SerInManual    = false;
+	  if (SerInBytes > 0) {
+	    SerInBuffer++;
+	    // More serial data available? If so, expect the next byte soon
+	    SerInBytes--;
+	  }
+	} else {
+	  // Automatic mode, receive through pokey.
+	  SerIn_Counter = 0;
+	  // Check whether SIO delivered any new bytes. If so,
+	  // generate an IRQ to deliver them.
+	  if (SerInBytes) {
+	    // Serial input just finished. Generate an IRQ.   
+	    GenerateIRQ(0x20);
+	  } else if (sio) {
+	    // Ask SIO again.
+	    sio->RequestInput();
+	  }
 	}
       }
+      if (update && SerInBytes && SerIn_Counter <= SerIn_Delay && SIOSound)
+	UpdateSound(0x0c);
     }
     //
     // Check whether there is some output from the serial register.
@@ -842,8 +933,8 @@ void Pokey::GoNSteps(int steps)
 	if (SerOut_BitCounter && --SerOut_BitCounter) {
 	  // Align the phase of timers 0 and 1 in two-tone mode.
 	  if (SkCtrl & 0x08) {
-	    Ch[0].DivNIRQ = 0;
-	    Ch[1].DivNIRQ = 0;
+	    Ch[0].DivNIRQ = Ch[0].DivNMax;
+	    Ch[1].DivNIRQ = Ch[1].DivNMax;
 	    Ch[0].DivNCnt = 0;
 	    Ch[1].DivNCnt = 0;
 	  }
@@ -895,51 +986,50 @@ void Pokey::GoNSteps(int steps)
       lastch = 2;
       // This locks timer 4 into reset. There is actually no
       // timer 3 interrupt, but for consisteny, reset it as well.
-      Ch[2].DivNIRQ = 0;
-      Ch[3].DivNIRQ = 0;
+      Ch[2].DivNIRQ = Ch[2].DivNMax;
+      Ch[3].DivNIRQ = Ch[3].DivNMax;
     } else {
       lastch = 4;
     }
     //
-    // Synchronization of channels for the two-tone modes.
-    if (SkCtrl & 0x08) { 
-      // Synchronized mode.
-      for(ch = 0, c = Ch;ch < lastch;ch++,c++) {
-	static const UBYTE irqbits[4] = {0x01,0x02,0x00,0x04}; 
-	// Yes, channel 3 cannot generate interrupts at all...
-	if ((c->DivNIRQ += steps) >= c->DivNMax) {
+    // Pokey timers. In reality, audio and IRQ are synchronous,
+    // they are not here - reason is that the audio timing is
+    // the real-world clock, but the IRQ timing is the emulated
+    // CPU clock.
+    for(ch = 0, c = Ch;ch < lastch;ch++,c++) {
+      // Yes, channel 3 cannot generate interrupts at all...
+      if ((c->DivNIRQ -= steps) <= 0) {
+	//
+	// Synchronized two-tone mode?
+	if (SkCtrl & 0x08) {
 	  switch(ch) {
 	  case 1:
 	    // Channel 1 syncs channel 0 always.
-	    Ch[0].DivNIRQ = 0;
+	    Ch[0].DivNIRQ = Ch[0].DivNMax;
 	    break;
 	  case 0:
 	    // Channel 0 syncs channel 1 if the serial register is set.
 	    if ((SerOut_Register & 0x01) && (SkCtrl & 0x80) == 0)
-	      Ch[1].DivNIRQ = 0;
+	      Ch[1].DivNIRQ = Ch[1].DivNMax;
 	    break;
 	  }
-	  //
-	  // Generate an IRQ now (or not for channel 3)
-	  if (IRQEnable & irqbits[ch]) {
-	    GenerateIRQ(irqbits[ch]);
-	  }	
-	  //c->DivNIRQ = 0; Allow accumulation of errors on average.
-	  c->DivNIRQ  -= c->DivNMax;
 	}
-      }
-    } else {
-      // Free-running mode.
-      for(ch = 0, c = Ch;ch < lastch;ch++,c++) {
-	static const UBYTE irqbits[4] = {0x01,0x02,0x00,0x04}; 
-	// Yes, channel 3 cannot generate interrupts at all...
-	if ((c->DivNIRQ += steps) >= c->DivNMax) {
-	  // Generate an IRQ now (or not for channel 3)
-	  if (IRQEnable & irqbits[ch]) {
-	    GenerateIRQ(irqbits[ch]);
-	  }	
-	  //c->DivNIRQ = 0; Allow accumulation of errors on average.
-	  c->DivNIRQ  -= c->DivNMax;
+	//
+	// Generate an IRQ now (or not for channel 3)
+	if (IRQEnable & irqbits[ch]) {
+	  GenerateIRQ(irqbits[ch]);
+	}	
+	//
+	// Reset the counters.
+	if ((AudioCtrl & LinkLoFlag[ch]) && c[1].DivNIRQ >= 0x100) {
+	  // This is the tricky case. The low-part of
+	  // the 16-bit counter with a partial underflow.
+	  c->DivNIRQ  += c->DivFullMax;
+	} else {
+	  // Average out errors by adding the timer constant
+	  // Equivalent if underflow to zero, otherwise its
+	  // the correct timing at least on average.
+	  c->DivNIRQ  += c->DivNMax;
 	}
       }
     }
@@ -975,6 +1065,53 @@ void Pokey::Step(void)
 }
 ///
 
+/// Pokey::VBI
+// VBI activity: This is only used here for dumping POKEY registers to generate
+// SAP files if this feature is enabled.
+void Pokey::VBI(class Timer *,bool,bool pause)
+{
+  if (pause == false && EnableSAP) {
+    // Check whether it makes sense to open up the SAP output.
+    if (SAPOutput == NULL && SongName && *SongName) {
+      bool enable = false;
+      int channel;
+      for(channel = 0;channel < 4;channel++) {
+	struct Channel *ch = Ch + channel;
+	if ((ch->AudioC & 0x0f) && (ch->AudioC & 0x10) == 0x00)
+	  enable = true;
+      }
+      if (enable) {
+	char filename[256];
+	snprintf(filename,sizeof(filename),"%s.sap",SongName);
+	SAPOutput = fopen(filename,"wb");
+	if (SAPOutput == NULL) {
+	  ThrowIo("Pokey::ParseArgs","unable to create the SAP output file");
+	} else {
+	  fprintf(SAPOutput,"SAP\r\n"
+		  "AUTHOR \"%s\"\r\n"
+		  "NAME \"%s\"\r\n"
+		  "TYPE R\r\n"
+		  "FASTPLAY %d\r\n"
+		  "\r\n",
+		  (AuthorName && *AuthorName)?AuthorName:("<?>"),
+		  SongName,
+		  NTSC?262:312);
+	}
+      }
+    }
+    if (SAPOutput) {
+      int channel;
+      for(channel = 0;channel < 4;channel++) {
+	struct Channel *ch = Ch + channel;
+	putc(ch->AudioF,SAPOutput);
+	putc(ch->AudioC,SAPOutput);
+      }
+      putc(AudioCtrl,SAPOutput);
+    }
+  }
+}
+///
+
 /// Pokey::HBI
 // Signal Pokey that a new scanline just
 // begun. This is not related to any real effect. 
@@ -1160,10 +1297,19 @@ UBYTE Pokey::RandomRead(void)
     // output, depending on AudCtrl.
     if (AudioCtrl & 0x80) {
       // The 9 bit poly counter.
-      return *Random9Ptr;
+      if (CycleTimers) {
+	return *Random9Ptr;
+      } else {
+	// Use also the CPU because otherwise the random generator is not very random...
+	return Random9[(Random9Ptr - Random9 + machine->CPU()->CurrentXPos()) % Poly9Size];
+      }
     } else {
       // The 17 bit poly counter.
-      return *Random17Ptr;
+      if (CycleTimers) {
+	return *Random17Ptr;
+      } else {
+	return Random17[(Random17Ptr - Random17 + machine->CPU()->CurrentXPos()) % Poly17Size];
+      }
     }
   } else {
     // This is also the first entry.
@@ -1194,7 +1340,27 @@ UBYTE Pokey::SerInRead(void)
 	// No else here, this is not redundant. The above might
 	// have caused a side effect!
 	if (SerInBytes > 0) {
-	  UBYTE byte = *SerInBuffer++;
+	  UBYTE byte;
+	  // Check whether the baud rate from the external source fits
+	  // the expectations. If not, read in something garbled.
+	  byte = *SerInBuffer++;
+	  if (SerInRate != SerialReceiveSpeed()) {
+	    // Ok, does not agree. Actually, for the tape the serial speed will
+	    // almost never agree. So check whether the speed varies "too much",
+	    // i.e. more than one bit in a byte, and fail then.
+	    int bitrate  = SerialReceiveSpeed();
+	    int recrate  = 20 * SerInRate;
+	    int byterate = 20 * bitrate;
+	    int delta    = recrate - byterate;
+	    if (delta < -bitrate || delta > bitrate) {
+	      // Difference is too large, fail
+	      do {
+		byte ^= UBYTE(rand() >> 8);
+	      } while(byte == 'A' || byte == 'C');
+	      // Just make sure that we don't generate a SIO 'ok' on error...
+	    }
+	  } 
+	  //
 	  // More serial data available? If so, expect the next byte soon
 	  SerInBytes--;
 	  // Check whether the input buffer is empty. If so, check whether
@@ -1288,19 +1454,21 @@ UBYTE Pokey::SkStatRead(void)
   // This emulates a device that operates at the 19200 baud the
   // SIO chain operates with, and does not recognize any internal
   // clock.
-  if (SerIn_Counter > 0 && SerInBuffer) {
-    const int timerconstant = (0x28 + 7); // This is the pokey timer constant for the 19200 baud control
-    int bitposition   = (SerIn_Counter + (timerconstant >> 1)) / timerconstant; // "half-bits"
+  if (SerIn_Counter > 0 && SerInBuffer && SerInRate > 0) {
+    int bitposition   = (SerIn_Counter + SerInRate - 1) / SerInRate; // "half-bits"
     bool bit          = true; // default level on the serial line is high.
     //
-    if ((LongStartBit && bitposition == 21) || bitposition == 20 || bitposition == 19) { 
+    if (bitposition == 20 || bitposition == 19) { 
       // The start bit. Actually, 1 1/2 start bits. If we assume that the
       // receiver waits 1 1/2 bits to sample at the middle of the bit, this
       // might be ok.
       bit = false;
     } else if (bitposition < 19 && bitposition > 2) {
       bit = ((*SerInBuffer >> (7 - ((bitposition - 3) >> 1))) & 0x01)?true:false;
-    } 
+    } else if (bitposition <= 2) {
+      // We are reading the stop bits. Consider this serial input as "missed".
+      SerInManual = true;
+    }
     //
     // Stop bits are one.
     //
@@ -1379,8 +1547,22 @@ void Pokey::STimerWrite(void)
   // This resets all counters to zero and
   // initializes the output flip-flops
   for(ch=0;ch < 4;ch++) {
-    Ch[ch].DivNCnt =  0;
-    Ch[ch].DivNIRQ =  -4;
+    Ch[ch].DivNCnt =  Ch[ch].DivNMax+4;
+    Ch[ch].DivNIRQ =  Ch[ch].DivNMax+4;
+  }
+  //
+  // Clock reset is offset if timers drive each
+  // other since the underrun requires some time
+  // to propagade.
+  if ((AudioCtrl & 0x50) == 0x50) {
+    // 1.79 Mhz clock on channel 0 also driving channel 1.
+    Ch[0].DivNCnt -= 3;
+    Ch[0].DivNIRQ -= 3;
+  }
+  if ((AudioCtrl & 0x28) == 0x28) {
+    // 1.79 Mhz clock on channel 2 also driving channel 3.
+    Ch[0].DivNCnt -= 3;
+    Ch[0].DivNIRQ -= 3;
   }
   //
   // STimer does not reset the polycounters
@@ -1406,6 +1588,9 @@ void Pokey::SerOutWrite(UBYTE val)
 	// Hacky. This is used for the concurrent mode.
 	// Transmit the byte thru the concurrent channel
 	sio->ConcurrentWrite(val);
+      } else if (SkCtrl & 0x08) {
+	// Two-tone mode enabled. Ok, write tape data.
+	sio->TapeWrite(val);
       } else {
 	// This is used for the regular SIO mode.
 	sio->WriteByte(val);
@@ -1541,7 +1726,7 @@ UBYTE Pokey::AllPotRead(void)
 // n 15Khz steps steps. Note that we might signal
 // zero bytes here in case we just want pokey to ask
 // back later...
-void Pokey::SignalSerialBytes(UBYTE *data,int num,UWORD delay)
+void Pokey::SignalSerialBytes(UBYTE *data,int num,UWORD delay,UWORD baudrate)
 {
   if (SerIn_Counter > 0 || SerInBytes) {
     // Serial problem of pokey: Serial device miscommunication,
@@ -1552,6 +1737,7 @@ void Pokey::SignalSerialBytes(UBYTE *data,int num,UWORD delay)
   }
   SerInBuffer    = data;
   SerInBytes     = num;
+  SerInRate      = baudrate;
   SerIn_Counter  = delay * Base15kHz + SerIn_Delay;
   //
   //
@@ -1605,6 +1791,23 @@ void Pokey::UpdateAudioMapping(void)
 }
 ///
 
+
+/// Pokey::SerialTransmitSpeed
+// Return the serial output speed as cycles of the 1.79Mhz clock.
+UWORD Pokey::SerialTransmitSpeed(void) const
+{
+  return SerOut_Clock;
+}
+///
+
+/// Pokey::SerialReceiveSpeed
+// Return the serial input speed as cycles of the 1.79Mhz clock.
+UWORD Pokey::SerialReceiveSpeed(void) const
+{ 
+  return SerIn_Clock;
+}
+///
+
 /// Pokey::DisplayStatus
 // Print the status of the chip over the monitor
 void Pokey::DisplayStatus(class Monitor *mon)
@@ -1628,8 +1831,8 @@ void Pokey::DisplayStatus(class Monitor *mon)
 		   "\tMax0    : %04x\tMax1    : %04x\tMax2    : %04x\tMax3      : %04x\n"
 		   "\tAudioCtrl : %02x\tSkStat    : %02x\tSkCtrl    : %02x\tKeyCode   : %02x\n"
 		   "\tIRQStat   : %02x\tIRQEnable : %02x\n"
-		   "\tSerInDly  : " LD "\tSerOutDly : " LD "\tSerXmtDly : " LD "\n"
-		   "\tSerInCnt  : " LD "\tSerOutCnt : " LD "\tSerXmtCnt : " LD "\n"
+		   "\tSerInDly  : " ATARIPP_LD "\tSerOutDly : " ATARIPP_LD "\tSerXmtDly : " ATARIPP_LD "\n"
+		   "\tSerInCnt  : " ATARIPP_LD "\tSerOutCnt : " ATARIPP_LD "\tSerXmtCnt : " ATARIPP_LD "\n"
 		   "\tSerInBytes: %d\tSerInData : %c%c\n",
 		   Unit,
 		   Ch[0].AudioF,Ch[1].AudioF,Ch[2].AudioF,Ch[3].AudioF,
@@ -1640,7 +1843,7 @@ void Pokey::DisplayStatus(class Monitor *mon)
 		   int((Ch[3].DivNIRQ > 0)?(Ch[3].DivNIRQ):(0)),
 		   int(Ch[0].DivNMax),int(Ch[1].DivNMax),int(Ch[2].DivNMax),int(Ch[3].DivNMax),
 		   AudioCtrl,SkStat   ,SkCtrl,KBCodeRead(),
-		    IRQStat  ,IRQEnable,
+		   IRQStatRead()      ,IRQEnable,
 		   SerIn_Delay,SerOut_Delay,SerXmtDone_Delay,
 		   SerIn_Counter,SerOut_Counter,SerXmtDone_Counter,SerInBytes,
 		   serin[0],serin[1]
@@ -1653,23 +1856,48 @@ void Pokey::DisplayStatus(class Monitor *mon)
 void Pokey::ParseArgs(class ArgParser *args)
 {  
   static const struct ArgParser::SelectionVector palvector[] = 
-    { {"PAL"          ,false},
-      {"NTSC"         ,true },
+    { {"Auto"         ,2    },
+      {"PAL"          ,0    },
+      {"NTSC"         ,1    },
       {NULL           ,0}
     };
   LONG ntsc;
-  bool cycle = CycleTimers;
+  bool cycle        = CycleTimers;
+  bool saprecording = EnableSAP;
   //
   //
-  ntsc = LONG(NTSC);
+  ntsc = NTSC?1:0;
+  if (isAuto)
+    ntsc = 2;
   args->DefineTitle((Unit)?"ExtraPokey":"Pokey");
   args->DefineLong("Volume","set Pokey volume in percent",0,300,Volume);
   args->DefineLong("Gamma","set Pokey output linearity in percent",50,150,Gamma);
-  args->DefineSelection("VideoMode","set POKEY base frequency",palvector,ntsc);
+  args->DefineSelection("PokeyTimeBase","set POKEY base frequency",palvector,ntsc);
   args->DefineBool("SIOSound","emulate serial transfer sounds",SIOSound);
   args->DefineBool("CyclePrecise","cycle precise pokey timers",cycle);
-  args->DefineBool("LongStartBit","enlarge the start bit to 1 1/2 bits",LongStartBit);
   args->DefineLong("FilterConstant","set high-pass filtering constant",0,1024,DCFilterConstant);
+  args->DefineBool("RecordAsSAP","record pokey output in a SAP file",saprecording);
+  //
+  // Potentially re-run the output.
+  if (saprecording != EnableSAP)
+    args->SignalBigChange(ArgParser::Reparse);
+  //
+  // Collect additional information in case SAP recording is enabled.
+  if (saprecording) {
+    args->DefineString("SAPName","name of the SAP song to record",SongName);
+    args->DefineString("SAPAuthor","author of the SAP song to record",AuthorName);
+  }
+  //
+  // Enable SAP.
+  if (saprecording) {
+    EnableSAP = true;
+  } else {
+    EnableSAP = false;
+    if (SAPOutput) {
+      fclose(SAPOutput);
+      SAPOutput = NULL;
+    }
+  }
   //
   // Initialize the pokey base clock.
   Frequency17Mhz     = (NTSC)?(1789773):(1773447); 
@@ -1689,9 +1917,16 @@ void Pokey::ParseArgs(class ArgParser *args)
     CycleTimers = cycle;
   }
   //
+  // For the auto selection, get the video mode from the machine.
+  if (ntsc == 2) {
+    ntsc   = machine->isNTSC();
+    isAuto = true;
+  } else {
+    isAuto = false;
+  }
+  //
   // If the video mode changed, ensure that the sound-subsystem is restarted.
   if (ntsc != LONG(NTSC)) {
-    NTSC = (ntsc)?(true):(false);
     args->SignalBigChange(ArgParser::Reparse);
   }
   // The global sound might have been redefined. Re-read here.
