@@ -2,7 +2,7 @@
  **
  ** Atari++ emulator (c) 2002 THOR-Software, Thomas Richter
  **
- ** $Id: x11_frontend.cpp,v 1.77 2009-08-10 20:19:06 thor Exp $
+ ** $Id: x11_frontend.cpp,v 1.90 2022/12/20 18:01:33 thor Exp $
  **
  ** In this module: A simple X11 frontend without further GUI
  **********************************************************************************/
@@ -17,7 +17,9 @@
 #include "gamecontroller.hpp"
 #include "exceptions.hpp"
 #include "x11_displaybuffer.hpp"
+#include "x11_xvideobuffer.hpp"
 #include "antic.hpp"
+#include "gtia.hpp"
 #include "stdio.hpp"
 #include "new.hpp"
 #ifndef X_DISPLAY_MISSING
@@ -30,6 +32,13 @@
 #define XK_XKB_KEYS
 #include <X11/keysymdef.h>
 #include <X11/cursorfont.h>
+#include "time.hpp"
+#ifndef HAS_STRUCT_TIMEVAL
+struct timeval {
+  int tv_sec;       /* Seconds.  */
+  int tv_usec;      /* Microseconds.  */
+};
+#endif
 ///
 
 /// ErrorHandler
@@ -55,11 +64,11 @@ static int ErrorHandler(Display *, XErrorEvent *ev)
 X11_FrontEnd::X11_FrontEnd(class Machine *mach,int unit)
   : XFront(mach,unit), visual(NULL), depth(0), defcolormap(0), custommap(0), 
     colormap(NULL), framebuffer(NULL), root(0),
-    isinit(false), ismapped(false), dump(false), grab(false), isgrabbed(false), havefocus(false), 
-    truecolor(false), showcursor(true), dumpcnt(1), scrolledlines(0),
+    isinit(false), ismapped(false), dump(false), isgrabbed(false), havefocus(false), 
+    truecolor(false), showcursor(true), button(false), dumpcnt(1), scrolledlines(0),
     LeftEdge(unit?0:16), TopEdge(0), 
     Width(unit?80*8:Antic::WindowWidth), Height(unit?25*8:Antic::WindowHeight),
-    PixelWidth(2), PixelHeight(2), PixMapIndirect(false),
+    PixelWidth(2), PixelHeight(2), PixMapIndirect(false), EnableXVideo(false),
     WMDeleteWindowAtom(0), WMProtocolsAtom(0), Shown(0), Hidden(0),
     KeypadStick(NULL), MouseStick(mach,"MouseStick",false), RelMouseStick(mach,"RelMouseStick",true), 
     PrivateCMap(false), SyncX(false), DisableDPMS(true),
@@ -91,10 +100,26 @@ class X11_DisplayBuffer *X11_FrontEnd::GetFrameBuffer(void)
 {
   if (framebuffer == NULL && isinit) { 
     truecolor   = machine->GTIA()->SuggestTrueColor();
-    framebuffer = XFront::FrameBufferOf(truecolor && depth > 8);
-    framebuffer->ConnectToX(display,screen,window,cmap,
-			    LeftEdge,TopEdge,Width,Height,
-			    PixelWidth,PixelHeight,PixMapIndirect);
+    framebuffer = XFront::FrameBufferOf(truecolor && depth > 8,EnableXVideo);
+    try {
+      framebuffer->ConnectToX(display,screen,window,cmap,
+			      LeftEdge,TopEdge,Width,Height,
+			      PixelWidth,PixelHeight,PixMapIndirect);
+    } catch(AtariException &aex) {
+      if (EnableXVideo) {
+	XFront::UnloadFrameBuffer();
+	framebuffer  = NULL;
+	EnableXVideo = false;
+	fixerror     = false;
+	framebuffer  = XFront::FrameBufferOf(truecolor && depth > 8,false);
+	framebuffer->ConnectToX(display,screen,window,cmap,
+				LeftEdge,TopEdge,Width,Height,
+				PixelWidth,PixelHeight,PixMapIndirect);
+      } else {
+	UnloadFrameBuffer();
+	throw;
+      }
+    }
   }
   return framebuffer;
 }
@@ -268,7 +293,8 @@ void X11_FrontEnd::CreateDisplay(void)
   // Now create the root window of this application
   //
   memset(&xswda,0,sizeof(xswda));
-  xswda.event_mask = KeyPressMask | KeyReleaseMask | ExposureMask | FocusChangeMask | ButtonPress;
+  xswda.event_mask = KeyPressMask | KeyReleaseMask | ExposureMask | FocusChangeMask | 
+    ButtonPress | ButtonRelease;
   xswda.colormap   = cmap;
   //
   window = XCreateWindow (display,root,0,0,emuwidth,emuheight,3,depth,
@@ -351,13 +377,38 @@ void X11_FrontEnd::CreateDisplay(void)
 #endif
   // This will build a new one. Must happen here in case the frame buffer
   // setup throws.
-  GetFrameBuffer();
+  try {
+    GetFrameBuffer();
+  } catch(AtariException &aex) {
+    UnloadFrameBuffer();
+    throw aex;
+  }
+  //
   XSync(display,false);    
   //XSetInputFocus(display,window,RevertToNone,CurrentTime);
   if (truecolor && depth <= 8) {
-    machine->PutWarning("Advanced image processing bypassed since "
+    machine->PutWarning("Advanced true color processing bypassed since "
 			"no true color display is available.");
   }
+  //
+  // If this is an xvideo frame buffer and we failed, try again. Otherwise,
+  // fail really!
+  if (fixerror) {
+    if (EnableXVideo) {
+      UnloadFrameBuffer();
+      EnableXVideo = false; 
+      // Sync and discard all events on the event queue
+      XSync(display,true);
+      fixerror     = false;
+      //
+      machine->PutWarning("XVideo display not available.");
+    } else {
+      UnloadFrameBuffer();
+      Throw(ObjectDoesntExist,"X11_FrontEnd::CreateDisplay",
+	    "Unable to create an X11 display.\n"
+	    "Sorry, no graphical output available.\n");
+    }
+  }
 }
 ///
 
@@ -566,6 +617,7 @@ void X11_FrontEnd::HandleKeyEvent(XKeyEvent *event)
     }
     break;
   case XK_BackSpace :
+  case XK_Terminate_Server:
     keyboard->HandleSimpleKey(downflag,0x08,shift,control); // backspace
     break;
   case XK_Delete:
@@ -613,6 +665,14 @@ void X11_FrontEnd::HandleButtonEvent(XButtonEvent *event)
     // A scroll-down event.
     scrolledlines++;
     break;
+  case 1:
+  case 2:
+    // Button events.
+    if (event->type == ButtonPress) {
+      button = true;
+    } else if (event->type == ButtonRelease) {
+      button = false;
+    }
   }
 }
 ///
@@ -767,19 +827,21 @@ void X11_FrontEnd::ShowPointer(bool showit)
 /// X11_FrontEnd::MousePosition
 // For GUI frontends within this buffer: Get the position and the status
 // of the mouse within this buffer measured in internal coordinates.
-void X11_FrontEnd::MousePosition(LONG &x,LONG &y,bool &button)
+void X11_FrontEnd::MousePosition(LONG &x,LONG &y,bool &but)
 {   
+  bool dummy;
+  //
   // Initialize the frontend if we do not have done so already
   if (!isinit) {
     CreateDisplay();
   } 
-  GetFrameBuffer()->MousePosition(x,y,button);
-  // If we don't have the focus, do not react on button presses
-  // This doesn't work reliable for twm, so disable.
-  /*
-  if (havefocus == false)
+  GetFrameBuffer()->MousePosition(x,y,dummy);
+  //
+  // Get the button state rather from the event chain and make sure
+  // we did not miss any mouse-up events.
+  if (dummy == false)
     button = false;
-  */
+  but = button;
 }
 ///
 
@@ -830,12 +892,45 @@ bool X11_FrontEnd::MouseMoveStick::TransmitStates(Display *d,Window w,
       button2 = (mask & Button3Mask)?true:false;
       //
       if (isrel) {
-	int dxl = (winx - lastx) << 12;
-	int dyl = (winy - lasty) << 12;
+	int dxl,dyl;
+	dxl = (winx - lastx) << 10;
+	dyl = (winy - lasty) << 10;
+#if HAVE_GETTIMEOFDAY
+	struct timeval tv;
+	if (gettimeofday(&tv,NULL) == 0) {
+	  int deltasec  = tv.tv_sec  - lastsec;
+	  int deltausec = tv.tv_usec - lastusec;
+	  
+	  lastsec  = tv.tv_sec;
+	  lastusec = tv.tv_usec;
+	  
+	  if (deltasec >= 0) {
+	    if (deltasec < (1L << 31) / (1000 * 1000)) {
+	      int deltatime = deltasec * 1000 * 1000 + deltausec;
+	      if (deltatime > 0) {
+		dxl = (((winx - lastx) << 24) / deltatime);
+		dyl = (((winy - lasty) << 24) / deltatime);
+	      }
+	    }
+	  }
+	}
+#endif	
 	//
 	lastx   = winx;
 	lasty   = winy;
 	//
+	// Avoid erratic non-movement due to no X-event on
+	// movement coming in.
+	if (dxl == 0 && dyl == 0 && (lastdx || lastdy)) {
+	  dxl = lastdx;
+	  dyl = lastdy;
+	  lastdx = 0;
+	  lastdy = 0;
+	} else {
+	  lastdx = dxl;
+	  lastdy = dyl;
+	}
+	//
 	dx      = (dxl > 32767)?(32767):((dxl < -32767)?(-32767):(dxl));
 	dy      = (dyl > 32767)?(32767):((dyl < -32767)?(-32767):(dyl));
 	//
@@ -995,6 +1090,7 @@ void X11_FrontEnd::ParseArgs(class ArgParser *args)
   LONG h        = Height;
   LONG format   = Format;
   bool indirect = PixMapIndirect;
+  bool xvideo   = EnableXVideo;
   static const struct ArgParser::SelectionVector Formats[] = {
     {"PNM", ScreenDump::PNM},
     {"BMP", ScreenDump::BMP},
@@ -1013,6 +1109,9 @@ void X11_FrontEnd::ParseArgs(class ArgParser *args)
   args->DefineBool("SyncX","enforce synchronous X rendering",SyncX);
 #ifdef USE_DPMS
   args->DefineBool("DisableDPMS","disable screen power saving",DisableDPMS);
+#endif
+#ifdef X_USE_XV
+  args->DefineBool("XVideoRendering","render through XVideo extension",EnableXVideo);
 #endif
   args->DefineBool("RenderIndirect","enable rendering thru a pixmap",PixMapIndirect);
   args->DefineString("ScreenBase","file base name for screen dumps",PictureBaseName);
@@ -1031,6 +1130,15 @@ void X11_FrontEnd::ParseArgs(class ArgParser *args)
   Format = (ScreenDump::GfxFormat)format;
 #ifdef USE_DPMS
   disableDPMS(display,DisableDPMS);
+  if (fixerror && xvideo) {
+    EnableXVideo = false;
+    UnloadFrameBuffer();
+    // Sync and discard all events on the event queue
+    XSync(display,true);
+    fixerror     = false; 
+    args->SignalBigChange(ArgParser::Reparse);
+    machine->PutWarning("XVideo display not available, disabling it.");
+  }
 #endif  
   if (pxwidth   != PixelWidth     ||
       pxheight  != PixelHeight    ||
@@ -1043,7 +1151,7 @@ void X11_FrontEnd::ParseArgs(class ArgParser *args)
     args->SignalBigChange(ArgParser::Reparse);
     CloseDisplay();
   }
-  if (truecolor != machine->GTIA()->SuggestTrueColor() ||
+  if (truecolor != machine->GTIA()->SuggestTrueColor() || xvideo != EnableXVideo   ||
       (colormap && (colormap != machine->GTIA()->ActiveColorMap()))) {
     args->SignalBigChange(ArgParser::Reparse);
     UnloadFrameBuffer();
@@ -1061,13 +1169,13 @@ void X11_FrontEnd::DisplayStatus(class Monitor *mon)
 		   "\tSyncX                 : %s\n"
 		   "\tScreen dump base name : %s\n"
 		   "\tIndirect rendering    : %s\n"
-		   "\tPixel width           : " LD "\n"
-		   "\tPixel height          : " LD "\n"
+		   "\tPixel width           : " ATARIPP_LD "\n"
+		   "\tPixel height          : " ATARIPP_LD "\n"
 		   "\tTrue color display    : %s\n"
-		   "\tLeftEdge              : " LD "\n"
-		   "\tTopEdge               : " LD "\n"
-		   "\tWidth                 : " LD "\n"
-		   "\tHeight                : " LD "\n",
+		   "\tLeftEdge              : " ATARIPP_LD "\n"
+		   "\tTopEdge               : " ATARIPP_LD "\n"
+		   "\tWidth                 : " ATARIPP_LD "\n"
+		   "\tHeight                : " ATARIPP_LD "\n",
 		   PrivateCMap?("on"):("off"),
 		   SyncX?("on"):("off"),
 		   PictureBaseName,		   
